
from copy import deepcopy

from magnum import *
from magnum.inverterdevice import InverterDevice
class RemoteDevice:

    noAGS = ["ampsstartdelay", "ampsstopdelay", "ampstart", "ampstop", "begintime", "cool", "exercisedays", "exerciseruntime", "exercisestart", "genstart", "maxrun", "quietbegintime", "quietendtime",
             "quiettime", "remotetimehours", "remotetimemins", "runtime", "socstart", "socstop", "starttemp", "startvdc", "stoptime", "topoff", "vdcstop", "voltstartdelay", "voltstopdelay", "warmup"]

    noBMK = ["batteryefficiency"]

    # noMSH = ["mshinputamps", "mshcutoutvoltage"]
    noPT100 = ["address", "packet", "lognumber", "relayonvdc", "relayoffvdc", "relayondelayseconds",
               "relaydelayoffseconds", "batterytempcomp", "powersavetime", "alarmonvdc",
               "alarmoffvdc", "alarmdondelay", "alarmoffdelay", "eqdonetimer", "chargerate",
               "rebulkonsunup", "AbsorbVoltage", "FloatVoltage", "EqualizeVoltage", "AbsorbTime",
               "RebulkVoltage", "BatteryTemperatureCompensation"]

    def __init__(self, trace=False):
        self.trace = trace
        self.data = {}
        self.deviceData = {}
        self.deviceData["device"] = REMOTE
        self.deviceData["data"] = self.data
        if self.trace:
            self.data["trace"]  = []
        self.data["revision"] = str("0.0")
        # self.data["action"] = 0
        self.data["searchwatts"] = 0
        self.data["batterysize"] = 0
        # see extract method for a note
        self.data["battype"] = 0
        self.data["absorb"] = 0
        self.data["chargeramps"] = 0
        self.data["ainput"] = 0
        self.data["parallel"] = 0
        # self.data["force_charge"] = 0
        # self.data["genstart"] = 0
        self.data["lbco"] = 0.0
        self.data["vaccutout"] = 0.0
        self.data["vsfloat"] = 0.0
        self.data["vEQ"] = 0.0
        self.data["absorbtime"] = 0
        # end of core info
        # A0
        self.data["remotetimehours"] = 0
        self.data["remotetimemins"] = 0
        self.data["runtime"] = 0
        self.data["starttemp"] = 0.0
        self.data["startvdc"] = 0.0
        self.data["quiettime"] = 0
        # A1
        self.data["begintime"] = 0
        self.data["stoptime"] = 0
        self.data["vdcstop"] = 0.0
        self.data["voltstartdelay"] = 0
        self.data["voltstopdelay"] = 0
        self.data["maxrun"] = 0
        # A2
        self.data["socstart"] = 0
        self.data["socstop"] = 0
        self.data["ampstart"] = 0.0
        self.data["ampsstartdelay"] = 0
        self.data["ampstop"] = 0
        self.data["ampsstopdelay"] = 0
        # A3
        self.data["quietbegintime"] = 0
        self.data["quietendtime"] = 0
        self.data["exercisedays"] = 0
        self.data["exercisestart"] = 0
        self.data["exerciseruntime"] = 0
        self.data["topoff"] = 0
        # A4
        self.data["warmup"] = 0
        self.data["cool"] = 0
        # 80
        self.data["batteryefficiency"] = 0
        # self.data["resetbmk"] = 0
        # 11 MSH - not supported
        # self.data["mshinputamps"] = 0
        # self.data["mshcutoutvoltage"] = 0
        # C0
        # self.data[" forcechgcode"] = 0
        # self.data["relayonoff"] = 0
        # self.data["buzzeronoff"] = 0
        # self.data["resetpt100"] = 0
        # self.data["address"] = 0
        # self.data["packet"] = 0
        # self.data["lognumber"] = 0
        # # C1
        # self.data["relayonvdc"] = 0
        # self.data["relayoffvdc"] = 0
        # self.data["relayondelayseconds"] = 0
        # self.data["relaydelayoffseconds"] = 0
        # self.data["batterytempcomp"] = 0
        # self.data["powersavetime"] = 0
        # # C2
        # self.data["alarmonvdc"] = 0
        # self.data["alarmoffvdc"] = 0
        # self.data["alarmdondelay"] = 0
        # self.data["alarmoffdelay"] = 0
        # self.data["eqdonetimer"] = 0
        # self.data["chargerate"] = 0
        # self.data["rebulkonsunup"] = 0
        # # C3
        # self.data["AbsorbVoltage"] = 0
        # self.data["FloatVoltage"] = 0
        # self.data["EqualizeVoltage"] = 0
        # self.data["RebulkVoltage"] = 0
        # self.data["BatteryTemperatureCompensation"] = 0

    def setBaseValues(self, unpacked):
        # self.data["action"] = unpacked[0]
        self.data["searchwatts"] = unpacked[1]
        value = unpacked[3]
        if(value > 100):
            self.data["absorb"] = value * InverterDevice.multiplier / 10
            self.data["battype"] = 0
        else:
            self.data["absorb"] = 0
            self.data["battype"] = value
        self.data["chargeramps"] = unpacked[4]
        self.data["ainput"] =  unpacked[5]
        self.data["revision"] = str(unpacked[6] / 10)
        value = unpacked[7]
        self.data["parallel"] = (value & 0x0f) * 10
        # self.data["force_charge"] = value & 0xf0
        # self.data["force_charge"] = self.data["force_charge"] >> 4
        # self.data["genstart"] = unpacked[8]
        self.data["lbco"] = unpacked[9] / 10
        self.data["vaccutout"] = float(unpacked[10])
        self.data["vsfloat"] = unpacked[11] * InverterDevice.multiplier / 10
        self.data["vEQ"] = self.data["absorb"] + (unpacked[12] / 10)
        self.data["absorbtime"] = unpacked[13] / 10

    def parse(self, packet):
        packetType = packet[0]
        unpacked = packet[2]
        if self.trace:
            self.data["trace"] .append((packetType, packet[1].hex().upper()))
            self.data["trace"]= list(set(self.data["trace"]))
        if packetType == REMOTE_C:
            self.setBaseValues(unpacked)
            #
            # The documentation is very weird on this value
            #
            self.data["batterysize"] = (unpacked[2] * 10)
            for key in ["vsfloat","vEQ","absorbtime", "remotetimehours","remotetimemins"]:
                self.data.pop(key, '')
        elif packetType == REMOTE_00:
            self.setBaseValues(unpacked)
        elif packetType == REMOTE_80:
            self.setBaseValues(unpacked)
            # batterysize is ill defined in documentation
            self.data["batterysize"] = (unpacked[2] * 10)
            self.data["remotetimehours"] = unpacked[14]
            self.data["remotetimemins"] = unpacked[15]
            self.data["batteryefficiency"] = unpacked[16]
            # self.data["resetbmk"] = unpacked[17]
        elif packetType == REMOTE_A0:
            self.setBaseValues(unpacked)
            self.data["remotetimehours"] = unpacked[14]
            self.data["remotetimemins"] = unpacked[15]
            self.data["runtime"] = unpacked[16] / 10
            self.data["starttemp"] = float(unpacked[17])
            self.data["starttemp"] = round((self.data["starttemp"] - 32) * 5 / 9, 1)
            value = unpacked[18] * InverterDevice.multiplier
            self.data["startvdc"] = value / 10
            self.data["quiettime"] = unpacked[19]
        elif packetType == REMOTE_A1:
            self.setBaseValues(unpacked)
            minutes = unpacked[14] * 15
            self.data["begintime"] = ((minutes // 60) * 100) + (minutes % 60)
            minutes = unpacked[15] * 15
            self.data["stoptime"] = ((minutes // 60) * 100) + (minutes % 60)
            value = unpacked[16]
            self.data["vdcstop"] = value * InverterDevice.multiplier / 10
            self.data["voltstartdelay"] = unpacked[17]
            if self.data["voltstartdelay"] > 127:
                self.data["voltstartdelay"] = (
                    self.data["voltstartdelay"] & 0x0f) * 60
            self.data["voltstopdelay"] = unpacked[18]
            if self.data["voltstopdelay"] > 127:
                self.data["voltstopdelay"] = (
                    self.data["voltstopdelay"] & 0x0f) * 60
            self.data["maxrun"] = unpacked[19] / 10
        elif packetType == REMOTE_A2:
            self.setBaseValues(unpacked)
            self.data["socstart"] = unpacked[14]
            self.data["socstop"] = unpacked[15]
            self.data["ampstart"] = float(unpacked[16])
            self.data["ampsstartdelay"] = unpacked[17]
            if self.data["ampsstartdelay"] > 127:
                self.data["ampsstartdelay"] = (self.data["ampsstartdelay"] & 0x0f) * 60
            self.data["ampstop"] = unpacked[18]
            self.data["ampsstopdelay"] = unpacked[19]
            if self.data["ampsstopdelay"] > 127:
                self.data["ampsstopdelay"] = (
                    self.data["ampsstopdelay"] & 0x0f) * 60
        elif packetType == REMOTE_A3:
            self.setBaseValues(unpacked)
            minutes = unpacked[14] * 15
            self.data["quietbegintime"] = (
                (minutes // 60) * 100) + (minutes % 60)
            minutes = unpacked[15] * 15
            self.data["quietendtime"] = (
                (minutes // 60) * 100) + (minutes % 60)
            minutes = unpacked[16] * 15
            self.data["exercisestart"] = (
                (minutes // 60) * 100) + (minutes % 60)
            self.data["runtime"] = unpacked[17] / 10
            self.data["topoff"] = unpacked[18]
        elif packetType == REMOTE_A4:
            self.setBaseValues(unpacked)
            self.data["warmup"] = unpacked[14]
            if self.data["warmup"] > 127:
                self.data["warmup"] = (self.data["warmup"] & 0x0f) * 60
            self.data["cool"] = unpacked[15]
            if self.data["cool"] > 127:
                self.data["cool"] = (self.data["cool"] & 0x0f) * 60
        elif packetType in [REMOTE_11, REMOTE_C0, REMOTE_C1, REMOTE_C2, REMOTE_C3, REMOTE_D0]:
                self.setBaseValues(unpacked)
                #
                # REMOTE_11, REMOTE_C0 through REMOTE_C3 and REMOTE_D0 are ignored
                #
    def cleanup(self, bmk, ags, pt100):
        if bmk == None:
            for item in self.noBMK:
                if item in self.data:
                    self.data.pop(item)

        if ags == None:
            for item in self.noAGS:
                if item in self.data:
                    self.data.pop(item)

        if pt100 == None:
            for item in self.noPT100:
                if item in self.data:
                    self.data.pop(item)
        # remove MSH as it's not supported - yet
        # for item in self.noMSH:
        #     if item in self.data:
        #         self.data.pop(item)

    def getDevice(self):
        device = deepcopy(self.deviceData)
        if self.trace:
            self.data["trace"]  = []
        return device



#
# Copyright (c) 2018-2025 Charles Godwin <magnum@godwin.ca>
#
# SPDX-License-Identifier:    BSD-3-Clause
#
#  BUILDINFO
__version__="2.0.7"
#
# names of devices
#
ACLD = "ACLD"
AGS = "AGS"
BMK = "BMK"
INVERTER = "INVERTER"
PT100 = "PT100"
REMOTE = "REMOTE"
RTR = "RTR"
UNKNOWN = "UNKNOWN"
#
#  Names of all the packet types
#
AGS_A1 = "AGS_A1"
AGS_A2 = "AGS_A2"

BMK_81 = "BMK_81"

INV = "INVERTER"
INV_C = "INVERTER_C"

PT_C1 = "PT_C1"
PT_C2 = "PT_C2"
PT_C3 = "PT_C3"

REMOTE_C = "REMOTE_C"
REMOTE_00 = "REMOTE_00"
REMOTE_11 = "REMOTE_11"
REMOTE_80 = "REMOTE_80"
REMOTE_A0 = "REMOTE_A0"
REMOTE_A1 = "REMOTE_A1"
REMOTE_A2 = "REMOTE_A2"
REMOTE_A3 = "REMOTE_A3"
REMOTE_A4 = "REMOTE_A4"
REMOTE_C0 = "REMOTE_C0"
REMOTE_C1 = "REMOTE_C1"
REMOTE_C2 = "REMOTE_C2"
REMOTE_C3 = "REMOTE_C3"
REMOTE_D0 = "REMOTE_D0"

RTR_91 = "RTR_91"

ACLD_D1 = "ACLD_D1"



#
# Copyright (c) 2018-2022 Charles Godwin <magnum@godwin.ca>
#
# SPDX-License-Identifier:    BSD-3-Clause
#
# This code is provided as an example of a JSON object
# run the program with --help for details of options.
#
import json
import signal
import sys
import time

from datetime import datetime, timezone

# from tzlocal import get_localzone

import magnum
from magnum.magnum import Magnum
from magnum.magparser import MagnumArgumentParser


def sigint_handler(signal, frame):
    print('Interrupted. Shutting down.')
    sys.exit(0)


def main():
    signal.signal(signal.SIGINT, sigint_handler)
    parser = MagnumArgumentParser(description="Magnum Data Dump", prog="magdump", fromfile_prefix_chars='@',
                                  epilog="Refer to https://github.com/CharlesGodwin/pymagnum for details")
    parser.add_argument("--device", "-d", nargs='+', default=f"{'/dev/ttyUSB0' if parser.isPosix else 'COM1'}",
                    help="Serial device name (default: %(default)s). You can specify more than one.")
    parser.add_argument("--interval", "-i", default=0, type=int, dest='interval',
                        help="Interval, in seconds, between dump records, in seconds. 0 means once and exit. (default: %(default)s)")
    parser.add_argument("--verbose", '-v', action="store_true", default=False,
                        help="Display options at runtime (default: %(default)s)")
    parser.add_argument("--pretty", action="store_true", default=False,
                        help="Show JSON in pretty format (default: %(default)s)")
    seldom = parser.add_argument_group("Seldom used")
    seldom.add_argument('--version', action='version',
                        version="%(prog)s Version:{}".format(magnum.__version__))
    seldom.add_argument("--packets", default=50, type=int,
                        help="Number of packets to generate in reader (default: %(default)s)")
    seldom.add_argument("--timeout", default=0.005, type=float,
                        help="Timeout for serial read (default: %(default)s)")
    seldom.add_argument("--trace", action="store_true", default=False,
                        help="Add most recent raw packet(s) info to data (default: %(default)s)")
    seldom.add_argument("--nocleanup", action="store_true", default=False, dest='cleanpackets',
                        help="Suppress clean up of unknown packets (default: False)")
    seldom.add_argument("--allinone", action="store_true", default=False,
                        help="Process data as a flat single row (default: %(default)s)")
    args = parser.magnum_parse_args()
    if hasattr(args, 'v1'): # a relic but not harmful
        args.allinone = True
    if args.verbose:
        print('Magnum Dump Version:{0}'.format(magnum.__version__))
        print("Options:{0}".format(str(args)
                                  .replace('Namespace(', '')
                                   .replace(')', '')
                                   .replace('[', '')
                                   .replace('\'', '')
                                   .replace(']', '')))
    magnumReaders = {}
    for device in args.device:
        try:
            magnumReader = Magnum(device=device, packets=args.packets, trace=args.trace,
                                  timeout=args.timeout, cleanpackets=args.cleanpackets)
            magnumReader.getDevices()  # test read to see if all's good
            magnumReaders[magnumReader.getComm_Device()] = magnumReader
        except Exception as e:
            print("{0} {1}".format(device, str(e)))
    if len(magnumReaders) == 0:
        print("Error: There are no usable devices connected.")
        exit(2)
    if args.interval != 0 and args.verbose == True:
        print("Dumping every:{1} seconds. Using: {0} ".format(
            list(magnumReaders.keys()), args.interval))
    if args.pretty:
        indent = 2
    else:
        indent = None
    while True:
        start = time.time()
        commdevices = []
        timestamp = datetime.now(timezone.utc).replace(microsecond=0).astimezone().isoformat()
        for comm_device, magnumReader in magnumReaders.items():
            try:
                devices = magnumReader.getDevices()
                if len(devices) != 0:
                    alldata = {}
                    alldata["datetime"] = timestamp
                    alldata["device"] = 'MAGNUM'
                    alldata['comm_device'] = comm_device
                    magnumdata = []
                    for device in devices:
                        data = {}
                        data["device"] = device["device"]
                        data["data"] = device["data"]
                        magnumdata.append(data)
                    alldata["data"] = magnumdata
                    if args.allinone:
                        alldata = magnumReader.allinone(alldata)
                    commdevices.append(alldata)
            except Exception as e:
                print("{0} {1}".format(comm_device, str(e)))
        if len(commdevices) == 1:
            outputdata = commdevices [0]
        else:
            outputdata = commdevices
        print(json.dumps(
            outputdata, indent=indent, ensure_ascii=True, allow_nan=True, separators=(',', ':')))
        if args.interval == 0:
            break
        interval = time.time() - start
        sleep = args.interval - interval
        if sleep > 0:
            time.sleep(sleep)


if __name__ == '__main__':
    main()



#
# Copyright (c) 2018-2022 Charles Godwin <magnum@godwin.ca>
#
# SPDX-License-Identifier:    BSD-3-Clause
#
#  This module tests the functionality of the serial interface
#  by attempting to read 50 packets of data and parsing them
#  The results are dumped to the terminal

import os
import signal
import sys
import time
import argparse

import magnum
from magnum.magnum import Magnum
from magnum.magparser import MagnumArgumentParser


class Logger(object):
    def __init__(self, logname=None):
        self.terminal = sys.stdout
        if logname:
            self.log = open(logname, 'w')
        else:
            self.log = None

    def write(self, message):
        self.terminal.write(message)
        if self.log:
            self.log.write(message)

    def flush(self):
        if self.log and not self.log.closed:
            self.log.flush()

    def close(self):
        if self.log:
            self.log.close()

def sigint_handler(signal, frame):
    print('Interrupted. Shutting down.')
    sys.exit(0)


def main():
    signal.signal(signal.SIGINT, sigint_handler)
    parser = MagnumArgumentParser(
        description="Magnum RS485 Reader Test", prog="magtest",
        epilog="Use `python -m serial.tools.list_ports` to identify devices. This program does NOT support use of options @file.")
    parser.add_argument("--device", "-d", default="all",nargs='+',
                        help="Serial device name (default: All serial devices.)")
    parser.add_argument("--packets", default=50, type=int,
                        help="Number of packets to generate in reader (default: %(default)s)")
    parser.add_argument("--timeout", default=0.005, type=float,
                        help="Timeout for serial read (default: %(default)s)")

    seldom = parser.add_argument_group("Seldom used")
    seldom.add_argument('--version', action='version',
                        version="%(prog)s Version:{}".format(magnum.__version__))
    seldom.add_argument("--trace", action="store_true", default=False,
                        help="Add most recent raw packet(s) info to data (default: %(default)s)")
    seldom.add_argument("--nocleanup", action="store_true", default=False, dest='cleanpackets',
                        help="Suppress clean up of unknown packets (default: False)")
    seldom.add_argument("--log", action="store_true", default=False,
                        help="Write copy of program output to log file in current directory (default: %(default)s)")
    seldom.add_argument("--flip", action="store_true", default=False,
                        #    help="Flips bit in RS485 data stream (default: %(default)s)")
                        help=argparse.SUPPRESS)
    args = parser.magnum_parse_args()
    if args.flip:
        args.cleanpackets = False
    logfile = os.path.join(os.getcwd(), "magtest_" + time.strftime("%Y-%m-%dT%H-%M-%S") + ".txt")
    if args.log:
        print(f"Output is being logged to {logfile}")
        sys.stdout = Logger(logname=logfile)
    print(f"Magnum Test Version:{magnum.__version__}")
    print(f"Options:{str(args)[10:-1]}")
    for comm_device in args.device:
        try:
            reader = Magnum(device=comm_device, packets=args.packets, trace=args.trace,
                            timeout=args.timeout, cleanpackets=args.cleanpackets, flip=args.flip)
            print(f"Testing:{comm_device}")

        except Exception as e:
            print(f"{comm_device} {str(e)}")
            continue
        try:
            start = time.time()
            packets = reader.getPackets()
            duration = time.time() - start
            unknown = 0
            formatstring = "Length:{0:2} {1:10}=>{2}"
            device_list = {}
            for item in [magnum.INVERTER, magnum.REMOTE, magnum.RTR, magnum.BMK, magnum.AGS, magnum.PT100, magnum.ACLD]:
                device_list[item] = 'NA'
            device_list[magnum.RTR] = False
            for packet in packets:
                if packet[0] == magnum.UNKNOWN:
                    unknown += 1
                if args.trace:
                    end = ' decode:'
                else:
                    end = '\n'
                print(formatstring.format(
                    len(packet[1]), packet[0], packet[1].hex().upper()), end=end)
                if args.trace:
                    print(*packet[2], end=' ')
                    print(packet[3])
                packetType = packet[0]
                if packetType in (magnum.INV, magnum.INV_C):
                    device_list[magnum.INVERTER] = True
                elif packetType in (magnum.REMOTE_C,
                                    magnum.REMOTE_00):
                    device_list[magnum.REMOTE] = True
                elif packetType == magnum.REMOTE_11:
                    pass  # ignored
                elif packetType == magnum.REMOTE_80:
                    device_list[magnum.REMOTE] = True
                    if device_list[magnum.BMK] != True:
                        device_list[magnum.BMK] = False
                elif packetType in (
                        magnum.REMOTE_A0,
                        magnum.REMOTE_A1,
                        magnum.REMOTE_A2,
                        magnum.REMOTE_A3,
                        magnum.REMOTE_A4):
                    device_list[magnum.REMOTE] = True
                    if device_list[magnum.AGS] != True:
                        device_list[magnum.AGS] = False
                elif packetType in (
                    magnum.REMOTE_C0,
                    magnum.REMOTE_C1,
                    magnum.REMOTE_C2,
                    magnum.REMOTE_C3
                ):
                    device_list[magnum.REMOTE] = True
                    if device_list[magnum.PT100] != True:
                        device_list[magnum.PT100] = False
                elif packetType == magnum.REMOTE_D0:
                    device_list[magnum.REMOTE] = True
                    if device_list[magnum.ACLD] != True:
                        device_list[magnum.ACLD] = False
                elif packetType == magnum.BMK_81:
                    device_list[magnum.BMK] = True
                elif packetType in (magnum.AGS_A1, magnum.AGS_A2):
                    device_list[magnum.AGS] = True
                elif packetType == magnum.RTR_91:
                    device_list[magnum.RTR] = True
                elif packetType in (magnum.PT_C1, magnum.PT_C2, magnum.PT_C3):
                    device_list[magnum.PT100] = True
                elif packetType == magnum.ACLD_D1:
                    device_list[magnum.ACLD] = True

            format1 = "Packets:{0} of {1} with {2} UNKNOWN, in {3:2.2f} seconds"
            format2 = "Packets:{0} in {3:2.2f} seconds"
            format = format1 if unknown > 0 else format2
            print(format.format(len(packets), args.packets, unknown, duration))
        # Analyze packets
            if len(packets) > unknown:
                for key, value in device_list.items():
                    if value == 'NA':
                        print(f"{key} not supported")
                    elif value == False:
                        print(f"{key} not connected")
                    else:
                        print(f"{key} Detected")
                if device_list[magnum.PT100] == True:
                    print(f"{magnum.PT100} has limited support, contact the author.")
                if device_list[magnum.ACLD] == True:
                    print(
                        "f{magnum.ACLD} is not currently supported, contact the author.")
            if args.log:
                print(f"Output was logged to {logfile}")
        except Exception as e:
            print(f"{reader.getComm_Device()} {str(e)}")
            print(f"Error detected attempting to read network data on {comm_device} - test failed.")
            continue

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
#
# Copyright (c) 2023 Charles Godwin <magnum@godwin.ca>
#
# SPDX-License-Identifier:    BSD-3-Clause
#
# This code is provided to print a MySQL schema definition from a JSON file generate by magdump
# run the program with --help for details of options.
#  mag2sql --help
#
#  magdump|mag2sql
#  or
#  magdump|mag2sql>[filename].sql
#
import sys
import json
import argparse


def main():
    parser = argparse.ArgumentParser(description="Magnum JSON to MySQL Generator", prog="mag2sql", fromfile_prefix_chars='@',
                                     epilog="Expects input to be JSON generated by magdump. Refer to https://github.com/CharlesGodwin/pymagnum for details")
    parser.add_argument("--input", "-i", default=None,
                        help="Input file name `None` implies stdin (default: %(default)s)")
    args = parser.parse_args()
    if args.input == None:
        inputjson = sys.stdin.read()
        alldata = json.loads(inputjson)
    else:
        f = open(args.input, "r")
        alldata = json.load(f)
    db_database = alldata['device']
    print(f"create schema if not exists {db_database};")
    print(f"use {db_database};")
    for item in alldata['data']:
        table = item['device']
        print(f"create table if not exists {table} (")
        keyfield = "datetime"
        print(f"\t`{keyfield}` datetime DEFAULT now(),")
        for variable, value in item['data'].items():
            if type(value) == float:
                dbtype = "float"
            elif type(value) == int:
                dbtype = "integer"
            elif type(value) == bool:
                dbtype = "integer(1)"
            else:
                dbtype = "varchar(25)"
            print(f"\t`{variable}` {dbtype} default NULL,")
        print(
            f"\tPRIMARY KEY (`{keyfield}`)\n\t) ENGINE=InnoDB;\n")

if __name__ == '__main__':
    main()




import math
from copy import deepcopy

from magnum import *
from magnum.inverterdevice import InverterDevice

class PT100Device:
    def __init__(self, trace=False):
        # self.trace = trace
        self.trace = True # force packet dump for now
        self.data = {}
        self.deviceData = {}
        self.deviceData["device"] = PT100
        self.deviceData["data"] = self.data
        if self.trace:
            self.data["trace"]  = []
        self.data['revision'] = '1.x' # from 0xC2
        # Start of C1
        self.data["address"] = 0
        self.data["mode"] = 0
        self.data["mode_text"] = ''
        self.data["mode_hex"] = ''
        self.data["regulation"] = 0
        self.data["regulation_text"] = ''
        self.data["fault"] = 0
        self.data["fault_text"] = ''
        self.data["battery"] = 0.0
        self.data["battery_amps"] = 0.0
        self.data["pv_voltage"] = 0.0
        self.data["charge_time"] = 0.0
        self.data["target_battery_voltage"] = 0.0
        self.data["relay_state"] = 1
        self.data["alarm_state"] = 0
        # check page 14 for fan on and Day
        self.data["fan_on"] = 0
        self.data["day"] = 0
        #
        self.data["battery_temperature"] = 0.0
        self.data["inductor_temperature"] = 0.0
        self.data["fet_temperature"] = 0.0
        # Start of C2
        # 2022-11-08 09:18:48 data only generated if a C2 packet is received
        # self.data["lifetime_kwhrs"] = math.nan
        # self.data["resettable_kwhrs"] = math.nan
        # self.data["ground_fault_current"] = math.nan
        # self.data["nominal_battery_voltage"] = 12.0
        # self.data["stacker_info"] = 0
        # self.data['dip_switches'] = '00000000'
        # self.data["model"] = ''
        # self.data["output_current_rating"] = math.nan
        # self.data["input_voltage_rating"] = math.nan
        # Start of C3
        # self.data["daily_kwh"] = math.nan
        # self.data["max_daily_pv_volts"] = math.nan
        # self.data["max_daily_pv_volts_time"] = math.nan
        # self.data["max_daily_battery_volts"] = math.nan
        # self.data["max_daily_battery_volts_time"] = math.nan
        # self.data["minimum_daily_battery_volts"] = math.nan
        # self.data["minimum_daily_battery_volts_time"] = math.nan
        # self.data["daily_time_operational"] = math.nan
        # self.data["daily_amp_hours"] = math.nan
        # self.data["peak_daily_power"] = math.nan
        # self.data["peak_daily_power_time"] = math.nan

    def parse(self, packet):
        packetType = packet[0]
        unpacked = packet[2]
        if self.trace:
            self.data["trace"] .append((packetType, packet[1].hex().upper()))
            self.data["trace"]= list(set(self.data["trace"]))
        address = unpacked[1] & 0X07
        if packetType == PT_C1 and address == 0:
            self.data['address'] = address
            #  skip header
            byte_value = unpacked[2]
            #  byte 2 assumes lower 4 bits and upper 4 bits
            self.data['mode'] = byte_value & 0x0F
            self.data['mode_hex'] = hex(unpacked[2]).upper()
            self.data['regulation'] = byte_value >> 4 & 0x0F
            #  byte 3
            byte_value = unpacked[3]
            self.data['fault'] = byte_value >> 3
            self.data['battery'] = unpacked[4] / 10
            self.data['battery_amps'] = unpacked[5] / 10
            self.data['pv_voltage'] = unpacked[6] / 10
            self.data['charge_time'] = unpacked[7] / 10
            byte_value = unpacked[8]
            self.data['target_battery_voltage'] = byte_value / 10 * InverterDevice.multiplier
            byte_value = unpacked[9]
            self.data['relay_state'] = byte_value & 0x01
            self.data['alarm_state'] = (byte_value >> 1) & 0x01
            self.data["fan_on"] = (byte_value >> 3) & 0x01
            self.data["day"] = (byte_value  >> 4) & 0x01

            byte_value = unpacked[10]
            if byte_value == 0X97 or byte_value == 0X98:
                self.data['battery_temperature'] = None # Short or open
            else:
                self.data['battery_temperature'] = float(byte_value)
            byte_value = unpacked[11]
            self.data['inductor_temperature'] = float(byte_value)
            byte_value = unpacked[12]
            self.data['fet_temperature'] = float(byte_value)
            modes = {
                1: "Float",
                2: "Bulk",
                3: "Absorb",
                4: "EQ"
            }
            if self.data['mode'] in modes:
                self.data['mode_text'] = modes[self.data['mode']]
            else:
                self.data['mode_text'] = "Unknown"
            regulations = {
                0: "Off",
                1: "Voltage",
                2: "Current",
                3: "Temperature",
                4: "Hardware",
                15: "MPPT"
                }
            if self.data['regulation']  in regulations:
                self.data['regulation_text'] = regulations[self.data['regulation']]
            else:
                self.data['regulation_text'] = "Unknown"
            faults = {
                0: "No Fault",
                1: "Input breaker Fault",
                2: "OPS Fault",
                3: "PV High Fault",
                4: "Battery High Fault",
                5: "BTS Shorted Fault",
                6: "FET Overtemp Fault",
                7: "Battery Overtemp Fault",
                8: "Over Current Fault",
                9: "Internal Phase Fault",
                10: "Open BTS Fault",
                11: "Internal Fault 1",
                12: "GFP Fault",
                13: "ARC Fault",
                14: "NTC Fault",
                15: "HW overtemp Fault",
                16: "Overyemp Fault",
                17: "USB Fault",
                20: "Stack Fault",
                21: "Stack warning",
                22: "Stack DIP Fault",
                23: "Stack DT Fault"
                }
            if self.data['fault'] in faults:
                self.data['fault_text'] = faults[self.data['fault']]
            else:
                self.data['fault_text'] = "unknown"
        elif packetType == PT_C2 and address == 0:
            self.data['lifetime_kwhrs'] = unpacked[2] * 10
            self.data['resettable_kwhrs'] = unpacked[3] / 10
            self.data['ground_fault_current'] = unpacked[4]
            byte_value = unpacked[5]
            self.data['stacker_info'] = byte_value >> 2
            byte_value = byte_value & 0x03
            if byte_value == 0:
                 self.data['nominal_battery_voltage']  = 12.0
            else:
                 self.data['nominal_battery_voltage']  = float(byte_value * 24.0)
            self.data['dip_switches'] = format(unpacked[6], '08b')
            self.data['revision'] = str(unpacked[7] / 10)
            # self.data['model'] = unpacked[7]
            self.data['model'] = 100 # That will have to do for now
            self.data['output_current_rating'] = unpacked[8]
            self.data['input_voltage_rating'] = unpacked[9] * 10
        # elif packetType == PT_C3 and address == 0:
        #     short_value = unpacked[1]
        #     self.data['daily_kwh'] = unpacked[2] / 10
        #     self.data['max_daily_pv_volts'] = unpacked[3]
        #     self.data['max_daily_pv_volts_time'] = unpacked[4] / 10
        #     self.data['max_daily_battery_volts'] = unpacked[5]
        #     self.data['max_daily_battery_volts_time'] = unpacked[6] / 10
        #     self.data['minimum_daily_battery_volts'] = unpacked[7]
        #     self.data['minimum_daily_battery_volts_time'] = unpacked[8] / 10
        #     self.data['daily_time_operational'] = unpacked[7] / 10
        #     self.data['daily_amp_hours'] = unpacked[10]
        #     self.data['peak_daily_power'] = unpacked[11]
        #     self.data['peak_daily_power_time'] = unpacked[12] / 10

    def getDevice(self):
        device = deepcopy(self.deviceData)
        if self.trace:
            self.data["trace"]  = []
        return device




import argparse
import shlex
import os
from serial.tools.list_ports import comports


class MagnumArgumentParser(argparse.ArgumentParser):
    isPosix = os.name != 'nt'
    def convert_arg_line_to_args(self, arg_line):
        return shlex.split(arg_line, True, self.isPosix)

    # This method cleans up device
    def magnum_parse_args(self):
        args = self.parse_known_args()[0]
        if hasattr(args, 'device'):
            if args.device == '' or args.device == None:
                args.device = []
            elif not isinstance(args.device, list):
                args.device = [args.device]
            serial_ports = []
            for device in comports(include_links=True):
                if self.isPosix:
                    serial_ports.append(device.device)
                else:
                    serial_ports.append(device.device.upper())
            devices = {}
            if len(args.device) == 0:
                args.device.append("/dev/ttyUSB0" if self.isPosix else "COM1")
            for name in args.device:
                name = name.strip("\"'")
                if name != "":
                    if name.lower() == "all":
                        for port in serial_ports:
                            devices[port] = "serial"
                    else:
                        if (self.isPosix and name in serial_ports) or (self.isPosix == False and name in serial_ports):
                            devices[name] = "serial"
                        else:
                            if name.startswith("!") and len(name)>1:
                                name = name[1:]
                            if os.path.exists(name):
                                devices["!"+name] = "file"
                            else:
                                self.error(f"option --device {name} is not available")
        if len(devices) == 0:
            self.error(f"No available devices")
        args.device = list(dict.fromkeys(devices))
        if hasattr(args, 'timeout'):
            if args.timeout < 0 or args.timeout > 1.0:
                self.error(
                    "option --timeout: Must be a number (float) between 0 and 1 second. i.e. 0.005")
        if hasattr(args, 'packets'):
            if args.packets < 1:
                self.error("option --packets: Must be greater than 0.")
        if hasattr(args, 'cleanpackets'):
            args.cleanpackets = not args.cleanpackets
        return args



from copy import deepcopy

from magnum import *

class ACLDDevice:
    def __init__(self, trace=False):
        # self.trace = trace
        self.trace = True # force packet dump
        self.data = {}
        self.deviceData = {}
        self.deviceData["device"] = ACLD
        self.deviceData["data"] = self.data
        if self.trace:
            self.data["trace"]  = []
        self.data["revision"] = str("0.0")

    def parse(self, packet):
        packetType = packet[0]
        unpacked = packet[2]
        if self.trace:
            self.data["trace"].append((packetType, packet[1].hex().upper()))
            self.data["trace"]=list(set(self.data["trace"]))
        # if packetType == ACLD_D1:
        #     pass

    def getDevice(self):
        device = deepcopy(self.deviceData)
        if self.trace:
            self.data["trace"]  = []
        return device



import argparse

parser = argparse.ArgumentParser(description="Process a list of strings.")
parser.add_argument('--files', nargs='+', help='A list of filenames' default=["xx"])

args = parser.parse_args()

print(f"Received files: {args.files}")




from copy import deepcopy

from magnum import *

class BMKDevice:
    def __init__(self, trace=False):
        self.trace = trace
        self.data = {}
        self.deviceData = {}
        self.deviceData["device"] = BMK
        self.deviceData["data"] = self.data
        self.data["revision"] = str(0.0)
        self.data["soc"] = 0
        self.data["vdc"] = 0.0
        self.data["adc"] = 0.0
        self.data["vmin"] = 0.0
        self.data["vmax"] = 0.0
        self.data["amph"] = 0
        self.data["amphtrip"] = 0.0
        self.data["amphout"] = 0.0
        self.data["Fault"] = 0
        self.data["Fault_Text"] = ""
        if self.trace:
            self.data["trace"]  = []

    def parse(self, packet):
        packetType = packet[0]
        unpacked = packet[2]
        if self.trace:
            self.data["trace"] .append((packetType, packet[1].hex().upper()))
            self.data["trace"]= list(set(self.data["trace"]))
        if packetType == BMK_81:
            self.data["soc"] = unpacked[1]
            self.data["vdc"] = round(unpacked[2] / 100, 2)
            self.data["adc"] = round(unpacked[3] / 10, 1)
            self.data["vmin"] = round(unpacked[4] / 100, 2)
            self.data["vmax"] = round(unpacked[5] / 100, 2)
            self.data["amph"] = unpacked[6]
            self.data["amphtrip"] = round(unpacked[7] / 10, 1)
            self.data["amphout"] = round(unpacked[8] * 100, 2)
            self.data["revision"] = str(round(unpacked[9] / 10, 2))
            self.data["Fault"] = unpacked[10]
            if self.data["Fault"] == 0:
                self.data["Fault_Text"] = "Reserved"
            elif self.data["Fault"] == 1:
                self.data["Fault_Text"] = "Normal"
            elif self.data["Fault"] == 2:
                self.data["Fault_Text"] = "Fault Start"

    def getDevice(self):
        device = deepcopy(self.deviceData)
        if self.trace:
            self.data["trace"]  = []
        return device



#
# Copyright (c) 2018-2023 Charles Godwin <magnum@godwin.ca>
#
# SPDX-License-Identifier:    BSD-3-Clause
#
# DO NOT SORT IMPORTS
#

import os
from struct import unpack
from time import sleep

from uptime import uptime

# from magnum import magnum
# import magnum
from magnum import *
from magnum.aclddevice import ACLDDevice
from magnum.agsdevice import AGSDevice
from magnum.bmkdevice import BMKDevice
from magnum.inverterdevice import InverterDevice
from magnum.pt100device import PT100Device
from magnum.remotedevice import RemoteDevice
from magnum.rtrdevice import RTRDevice

# MAGNUM_DELAY can be defined in environment
# export MAGNUM_DELAY=nnn.nn

try:
    MAGNUM_DELAY = float(os.getenv('MAGNUM_DELAY', 30.0))
except:
    MAGNUM_DELAY = 30.0
# delay serial startup until system settles down - at least MAGNUM_DELAY seconds
delay = MAGNUM_DELAY - uptime() # type: ignore
if delay > 0.0:
    sleep(delay)
# This must be after the sleep(delay)
import serial  # noqa

class Magnum:
    '''
    :param device: The serial device to connect to, defaults to /dev/ttyUSB0
    :type device: str, optional
    :param packets: How many packets to capture in one sample, defaults to 50
    :type packets: int, optional
    :param cleanpackets: Allow system to try to fixup adjacent packets by merging them, defaults to True
    :type cleanpackets: boolean, optional
    :param timeout: How much time delay, in fractions of second, to trigger end of packet, defaults to 0.001 second
    :type timeout: float, optional
    :param trace: Enable adding last of every packet type processed. The packets, as HEX strings, are appended to data object, Defaults to False
    :type trace: boolean, optional
    '''

    sevenzeros = bytes([0, 0, 0, 0, 0, 0, 0])
    #
    # refer to struct.unpack for format description
    #
    default_remote = 'BBBBBbBBBBBBBB'
    unpackFormats = {
        AGS_A1: 'BbBbBB',
        AGS_A2: 'BBBHB',
        BMK_81: 'BbHhHHhHHBB',
        INV:   'BBhhBBBBBBBBBBBBhb',
        INV_C: "BBhhBBBBBBBBBB", # old style inverter
        PT_C1: 'BBBBHhHBBBBBB',
        PT_C2: 'BBHHBBBBBBB',
        PT_C3: '15B',  # double check documentation
        REMOTE_C:  default_remote + 'BB', # old style remote
        REMOTE_00: default_remote + '7B',
        REMOTE_11: default_remote + '7B', # just the first 2 bytes count
        REMOTE_80: default_remote + 'bbbb3B',
        REMOTE_A0: default_remote + 'BBBbBBB',
        REMOTE_A1: default_remote + '7B',
        REMOTE_A2: default_remote + 'bb5B',
        REMOTE_A3: default_remote + '7B',
        REMOTE_A4: default_remote + '7B',
        REMOTE_C0: default_remote + '7B',  # PT100  not used
        REMOTE_C1: default_remote + '7B',  # PT100  not used
        REMOTE_C2: default_remote + '7B',  # PT100  not used
        REMOTE_C3: default_remote + '7B',  # PT100  not used
        REMOTE_D0: default_remote + '7B',  # ACLD Needs work
        ACLD_D1: '7B',                     # ACLD Needs work
        RTR_91: 'BB',
        UNKNOWN: ''
    }

    def __init__(self, device="/dev/ttyUSB0", timeout=0.005, packets=50, cleanpackets=True, trace=False, flip=False):
        self.packetcount = packets
        self.timeout = timeout
        self.cleanpackets = cleanpackets
        self.trace = trace
        self.flip = flip
        self.reader = None
        self.inverter = None
        self.remote = None
        self.bmk = None
        self.ags = None
        self.rtr = None
        self.pt100 = None
        self.acld = None
        self.inverter_revision = -1
        self.inverter_model = -1
        if device.startswith("!"):
            self.comm_device = device[1:]
            self.stored_packets = self._load_packets(device)
        else:
            self.stored_packets = None
            self.comm_device = device
        return

    def getComm_Device(self):
        return self.comm_device

    def _load_packets(self, filename):
        packets = []
        with open(filename) as file:
            lines = file.readlines()
        for line in lines:
            line = line.strip()
            ix = line.find("#")
            if ix >= 0:
                line = line[0:ix].strip()
            ix = line.find("=>")
            if ix >= 0:
                line = line[ix+2:]
                ix = line.find(' ')
                if ix >= 0:
                    line = line[0:ix]
                data = bytes.fromhex(line)
                packets.append(data)
        if len(packets) == 0:
            raise ValueError(f"There were no valid records in {filename}")
        return packets
    #
    # returns a list of tuples of {message type, bytes of packet, and tuple of {unpacked packet values)}
    #

    def getPackets(self):
        '''
        Retrieves the raw packets. This is not normally used.

        :return: List of `tuple` objects
        :rtype: list

        **tuple contents**:

        - name of packet
        - bytes of packet
        - tuple of unpacked values - Based on ME documentation
        '''
        packets = self.readPackets()
        messages = []
        unknown = 0
        for packet in packets:
            message = self._parsePacket(packet)
            if message[0] == UNKNOWN:
                unknown += 1
            messages.append(message)
        #
        # if there at least 2 UNKNOWN packets
        # attempt to clean them up
        #
        if unknown > 1 and self.cleanpackets:
            messages = self.cleanup(messages)
        return messages
        #

    #  raw read of packets to bytes[]
    #

    def readPackets(self):
        packet = bytearray()
        packets = []
        if self.stored_packets != None:
            for ix in range(self.packetcount):
                packet = self.stored_packets.pop(0)
                packets.append(packet)
                self.stored_packets.append(packet)
            return packets
        if self.reader == None:
            self.reader = serial.serial_for_url(self.comm_device,
                                                baudrate=19200,
                                                bytesize=8,
                                                timeout=self.timeout,
                                                stopbits=serial.STOPBITS_ONE,
                                                dsrdtr=False,
                                                parity=serial.PARITY_NONE)
            self.reader.close()

        #
        # open port every time
        #
        self.reader.open()
        #
        # wait to see if there is any traffic on the device
        #
        sleep(0.25)
        if self.reader.in_waiting == 0:
            self.reader.close()
            self.reader = None
            raise ConnectionError("There doesn't seem to be a network")
        packetsleft = self.packetcount
        self.reader.reset_input_buffer()
        #
        # Start of packet reads into a list of bytearray()
        # This is a tight loop
        #
        while packetsleft > 0:
            readbytes = self.reader.read(self.reader.in_waiting or 1)
            packet += readbytes
            #
            # assumes an empty read is an inter packet gap
            #
            if len(readbytes) == 0 and len(packet) != 0:
                packets.append(packet)
                packetsleft -= 1
                packet = bytearray()
        self.reader.close()
        return packets
    #
    #
    # based on what we know from ME documentation
    # attempt to build a known packet and unpack its data into values
    #

    def _parsePacket(self, packet):
        # Needs work 2023-12-28
        if self.flip and len(packet) > 0:
            new_packet = []
            for i in range(len(packet)):
                new_packet.append((~packet[i]) & 0xff)
            packet = bytearray(new_packet)
        if len(packet) == 22:
            packet = packet[:21]
        elif len(packet) == 17:  # takes care of classic
            packet = packet[:16]
        packetType = UNKNOWN
        if len(packet) > 0:
            packetLen = len(packet)
            firstbyte = packet[0]
            lastbyte = packet[-1]
            if packetLen == 2 and firstbyte == 0x91:
                packetType = RTR_91
            elif packetLen == 6:
                if firstbyte == 0xa1:
                    packetType = AGS_A1
                elif firstbyte == 0xa2:
                    packetType = AGS_A2
            elif packetLen == 8 and firstbyte == 0xD1:
                packetType = ACLD_D1
            elif packetLen == 13 and firstbyte == 0xC2:
                packetType = PT_C2
            elif packetLen == 14 and firstbyte == 0xC3:
                packetType = PT_C3
            elif packetLen == 16:
                if firstbyte == 0xC1:
                    packetType = PT_C1
                elif packet[10] <= 0x27 and packet[14] in InverterDevice.inverter_models:
                    packetType = INV_C
                    if self.inverter_revision == -1:
                        self.inverter_revision = packet[10]
                        self.inverter_model = packet[14]
                else:
                    packetType = REMOTE_C
            elif packetLen == 18 and firstbyte == 0x81:
                packetType = BMK_81
            elif packetLen == 21:
                version = packet[10]
                model = packet[14]
                if lastbyte == 0 and firstbyte == 0:
                    #
                    #  There is an undocumented Remote message generated with seven 0x00 bytes at the end.
                    #  This code distinguishes it from a Inverter record with status byte 0 == 0x0
                    #
                    #  Also the ME-ARC sends a spurious record with a zero end byte
                    #
                    if packet[-7:] == self.sevenzeros:
                        packetType = REMOTE_00
                    else:
                        if version == (self.inverter_revision and model == self.inverter_model) or self.inverter_revision == -1:
                            packetType = INV
                        else:
                            packetType = REMOTE_00
                else:
                    if lastbyte == 0:
                        if (version == self.inverter_revision and model == self.inverter_model) or self.inverter_revision == -1:
                            if model in InverterDevice.inverter_models:
                                packetType = INV
                                if self.inverter_revision == -1:
                                    self.inverter_revision = version
                                    self.inverter_model = model
                        else:
                            packetType = REMOTE_00
                    elif lastbyte == 0xa0:
                        packetType = REMOTE_A0
                    elif lastbyte == 0xa1:
                        packetType = REMOTE_A1
                    elif lastbyte == 0xa2:
                        packetType = REMOTE_A2
                    elif lastbyte == 0xa3:
                        packetType = REMOTE_A3
                    elif lastbyte == 0xa4:
                        packetType = REMOTE_A4
                    elif lastbyte == 0x80:
                        packetType = REMOTE_80
                    elif lastbyte == 0xC0:
                        packetType = REMOTE_C0
                    elif lastbyte == 0xC1:
                        packetType = REMOTE_C1
                    elif lastbyte == 0xC2:
                        packetType = REMOTE_C2
                    elif lastbyte == 0xC3:
                        packetType = REMOTE_C3
                    elif lastbyte == 0x11:
                        packetType = REMOTE_11
                    elif lastbyte == 0xD0:
                        packetType = REMOTE_D0
            #
            # Unpack as big endian
            # Refer to unpackFormats
            #
            mask = ">" + self.unpackFormats[packetType]
            if len(mask) > 1:
                try:
                    fields = unpack(mask, packet)
                except Exception as e:
                    msg = "{0} Converting {1} - {2} bytes".format(
                        e.args[0], packetType, len(packet))
                    fields = {}
                    print(msg)
                    packetType = UNKNOWN
                    # raise unpack_error(msg) from e
            else:
                fields = {}
            return([packetType, packet, fields, self.unpackFormats[packetType]])

    #
    # cleanup looks for consecutive UNKNOWN packet pairs and concatenates the pair
    # and attempts to parse the result. It has reasonable success.
    #
    def cleanup(self, messages):
        cleaned = []
        lastone = len(messages) - 2
        ignoreit = False
        for index, message in enumerate(messages):
            if ignoreit:
                ignoreit = False
            elif index > lastone or message[0] != UNKNOWN:
                ignoreit = False
                cleaned.append(message)
            else:
                nextmessage = messages[index + 1]
                if nextmessage[0] == UNKNOWN:
                    # we may have a match
                    newmessage = self._parsePacket(message[1] + nextmessage[1])
                    ignoreit = True
                    cleaned.append(newmessage)
        return cleaned
    #
    #     returns an array of device ordered dictionary
    #
    #  Each class is instantiated only once per run time execution
    #  This allows an object to reflect the latest CUMULATIVE value for the packets.
    #  This is useful for PT100 and AGS packets which are not too numerous
    #
    #  returns a deepcopy of the device data collections
    #

    def getDevices(self):
        '''
        Get a list of connected devices

        :return: List of device dictionaries
        :rtype: list

        Each dictionary has two or, optionally, three items:

        - **device**  One of INVERTER, REMOTE, AGS, BMK or PT100
        - **data** A dictionary of name/value pairs for the device.
        - **trace** If trace is set to True then trace will have a list of tupples of every packet since last time invoked
        '''
        # pass each the packets to the correct object
        #
        # each packet is a tuple of:
        #     type(string) name of packet
        #     raw packet (bytes) the raw binary bytes of the packet
        #     unpacked data (tuple int) integers of data deconstructed to match ME documentation
        #
        for packet in self.getPackets():
            packetType = packet[0]
            if packetType in (INV, INV_C):
                if self.inverter == None:
                    self.inverter = InverterDevice(trace=self.trace)
                self.inverter.parse(packet)
            elif packetType in (REMOTE_C,
                                REMOTE_00,
                                REMOTE_11,
                                REMOTE_80,
                                REMOTE_A0,
                                REMOTE_A1,
                                REMOTE_A2,
                                REMOTE_A3,
                                REMOTE_A4,
                                REMOTE_C0,
                                REMOTE_C1,
                                REMOTE_C2,
                                REMOTE_C3,
                                REMOTE_D0):
                if self.remote == None:
                    self.remote = RemoteDevice(trace=self.trace)
                self.remote.parse(packet)
            elif packetType == BMK_81:
                if self.bmk == None:
                    self.bmk = BMKDevice(trace=self.trace)
                self.bmk.parse(packet)
            elif packetType in (AGS_A1, AGS_A2):
                if self.ags == None:
                    self.ags = AGSDevice(trace=self.trace)
                self.ags.parse(packet)
            elif packetType == RTR_91:
                if self.rtr == None:
                    self.rtr = RTRDevice(trace=self.trace)
                self.rtr.parse(packet)
            elif packetType in (PT_C1, PT_C2, PT_C3):
                if self.pt100 == None:
                    self.pt100 = PT100Device(trace=self.trace)
                self.pt100.parse(packet)
            elif packetType == ACLD_D1:
                if self.acld == None:
                    self.acld = ACLDDevice(trace=self.trace)
                self.acld.parse(packet)
        if self.remote:
            #
            # remove extraneous REMOTE fields if corresponding device is not present
            #
            self.remote.cleanup(self.bmk, self.ags, self.pt100)
        devices = []
        for device in [self.inverter, self.remote, self.bmk, self.ags, self.rtr, self.pt100, self.acld]:
            if device:
                deviceinfo = device.getDevice()
                if deviceinfo:
                    devices.append(deviceinfo)
        return devices

# 2023-11-08 15:22:24 Added

# This merges all device data into one long dictionary. Each variable is prefixed with device name
# consistent with the values used in the old Java based magnum software (v1)
#
    def allinone(self, devices):
        returndata = []

        deviceprefixes = {"INVERTER": "INV",
                        "AGS": "AGS",
                        "BMK": "BMK",
                        "RTR": "RTR",
                        "PT100": "PT",
                        "REMOTE": "ARC",
                        "RTR": "RTR"
                        }

        if (type(devices) != list):
            devices = [devices]
        for device in devices:
            newdata = {}
            newdata['datetime'] = device['datetime']
            newdata['comm_device'] = device['comm_device']
            newdata['device'] = device["device"]
            data = []
            itemdata={}
            for item in device['data']:
                if item['device'] in deviceprefixes:
                    deviceprefix = deviceprefixes[item['device']]
                else:
                    deviceprefix = item['device']
                itemstuff = item['data']
                for itemkey, itemvalue in itemstuff.items():
                    key = f"{deviceprefix}_{itemkey}"
                    itemdata[key] = itemvalue
            newblock = {}
            newblock['device'] = "log_data"
            newblock['data'] = dict(sorted(itemdata.items()))
            newdata['data'] = [newblock]
            returndata.append(newdata)
        if len(returndata) == 1:
            return returndata[0]
        else:
            return returndata
